class SprinklerException(Exception):
    """
    All Sprinkler exceptions inherit from this.
    """


class TargetsDirectoryUnsetException(SprinklerException):
    """
    raised when an action is taken that requires the TARGETS_DIRECTORY
    to be set, but it is not
    """


class GitException(SprinklerException):
    """
    Raised when there is an exception with a git command.
    """


class APIKeyAlreadyExistsException(SprinklerException):
    """
    raised when creating an API Key that already exists.
    """


class APIKeyDoesNotExistException(SprinklerException):
    """
    raised when an action is taken on an API key that does not exist.
    """


class InvalidConfigKey(SprinklerException):
    """
    raised when an operation is attempted on a non-existent config key.
    """


class TargetBuildException(SprinklerException):
    """
    raised when there is an exception while building a target.
    """


class BuildLogNotFoundException(SprinklerException):
    """
    raised when build logs are not found for a target.
    """


class TaskInstanceDoesNotExistException(SprinklerException):
    """
    Raised when a task instance does not exist in the db.
    """


class InvalidTaskInstanceQueryFilter(SprinklerException):
    """
    Raised when an invalid filter is provided
    """


class UserAlreadyExistsException(SprinklerException):
    """
    raised when creating a user that already exists.
    """


class UserDoesNotExistException(SprinklerException):
    """
    raised when an action is taken on a user that doesn't exist.
    """


class InvalidSprinklerConfigException(SprinklerException):
    """
    Exception raised when the sprinkler_config.yaml is invalid.
    """


class TaskCreationException(SprinklerException):
    """
    raised when a task is unable to be created.
    """


class InvalidSecretName(SprinklerException):
    """
    raised when a secret name is invalid.
    """



from typing import Literal

_CONFIG_YAML_FILENAME = "sprinkler_config.yaml"

_TASK_MANAGER_LOGGER = "sprinkler.task_manager"
_TASK_INSTANCE_LOGGER = "sprinkler.task_instance"
_SCHEDULER_LOGGER = "sprinkler.scheduler"
_WEBSERVER_LOGGER = "sprinkler.webserver"
_IMAGES_LOGGER = "sprinkler.images"
_SECRETS_LOGGER = "sprinkler.secrets"
_GIT_LOGGER = "sprinkler.git"

_SPRINKLER_BODY_ENV_VAR = "SPRINKLER_TASK_INSTANCE_BODY"

_TASK_TYPE_FIRE_AND_FORGET = "fire-and-forget"
_TASK_TYPE_CALL_AND_RESPONSE = "call-and-response"
TASK_TYPE = Literal["fire-and-forget", "call-and-response"]

_TRIGGER_TYPE_SCHEDULED = "scheduled"
_TRIGGER_TYPE_WEBHOOK = "webhook"

_TASK_INSTANCE_STATUS_WAITING = "waiting"
_TASK_INSTANCE_STATUS_RUNNING = "running"
_TASK_INSTANCE_STATUS_ERROR = "error"
_TASK_INSTANCE_STATUS_COMPLETE = "complete"
_TASK_INSTANCE_STATUS_TIMEOUT = "timeout"

_TASK_EXECUTE_LOOP_DELAY_SLOW = 0.5
_TASK_EXECUTE_LOOP_DELAY_FAST = 0.05

TRIGGER_TYPE = Literal["scheduled", "webhook"]
TASK_INSTANCE_STATUS_TYPE = Literal[
    "waiting", "running", "error", "complete", "timeout"
]
_EXECUTE_TASK_INSTANCE_CHANNEL = "sprinkler_execute_task_instance"

_SECRET_ENVIRONMENT_VARIABLE_PREFIX = "SPRINKLER_SECRET_"

_SPRINKLER_SECRET_KEY_ENVIRONMENT_VARIABLE = "SPRINKLER_SECRET"

RUNTIME_TYPE = Literal["python", "node-npm", "node-yarn"]



import os
import sqlite3
from typing import Optional, Tuple

from .utils import get_sprinkler_directory


def get_db_file():
    return os.path.join(get_sprinkler_directory(), "sprinkler.db")


def get_db():
    return sqlite3.connect(get_db_file())


def db_select(conn: sqlite3.Connection, query: str, args: Optional[Tuple]) -> list:
    cursor = conn.cursor()
    if args is None:
        cursor.execute(query)
    else:
        cursor.execute(query, args)
    results = cursor.fetchall()
    cursor.close()
    return results


def db_execute(conn: sqlite3.Connection, query: str, args: Optional[Tuple]) -> None:
    cursor = conn.cursor()
    if args is None:
        cursor.execute(query)
    else:
        cursor.execute(query, args)
    cursor.close()
    conn.commit()


DDL = [
    """
CREATE TABLE IF NOT EXISTS task_instances (
    id TEXT,
    target TEXT,
    task TEXT,
    timeout INTEGER,
    entrypoint_file TEXT,
    task_type TEXT,
    status TEXT,
    trigger_type TEXT,
    created_at_time REAL,
    runtime TEXT,
    version TEXT,
    git_hash TEXT,

    start_time REAL,
    end_time REAL,
    log_location TEXT,
    container_id TEXT,
    response_channel TEXT,
    is_failure_task BOOL,


    primary key (id)
);
    """.strip(),
    """
CREATE TABLE IF NOT EXISTS secrets (
    key TEXT,
    value TEXT,
    primary key (key)
);
    """.strip(),
    """
CREATE TABLE IF NOT EXISTS config (
    key TEXT,
    value TEXT,
    primary key (key)
);
    """.strip(),
    """
CREATE TABLE IF NOT EXISTS target_build_logs (
    target TEXT,
    build_logs TEXT,
    success BOOLEAN,
    primary key (target)
);
    """.strip(),
    """
CREATE TABLE IF NOT EXISTS users (
    user_name TEXT,
    password_hash TEXT,
    admin BOOLEAN,
    primary key (user_name)
);
    """.strip(),
    """
CREATE TABLE IF NOT EXISTS api_keys (
    api_key_id TEXT,
    api_key_hash TEXT,
    primary key (api_key_id)
);
    """.strip(),
]


def init_db():
    conn = get_db()
    cursor = conn.cursor()
    for table in DDL:
        cursor.execute(table)
        conn.commit()
    cursor.close()
    conn.close()



import logging
import os
import shutil
import sqlite3
from typing import List, Optional, Tuple
from uuid import uuid4

import docker
import docker.errors

from ..constants import _IMAGES_LOGGER, RUNTIME_TYPE
from ..db import db_execute, db_select
from ..exceptions import BuildLogNotFoundException, TargetBuildException
from ..targets.sprinkler_config import load_targets
from ..utils import make_target_image_name

logger = logging.getLogger(_IMAGES_LOGGER)


class Images:
    db: sqlite3.Connection
    docker_client: docker.DockerClient

    def __init__(self, db: sqlite3.Connection, docker_client: docker.DockerClient):
        self.db = db
        self.docker_client = docker_client

    def build_target_images(
        self,
        targets_dir: str,
    ):
        targets = load_targets(targets_dir, include_errors=False)
        for target, config in targets:
            self.build_target_image(
                targets_dir,
                target,
                config.runtime,
                config.version,
                config.os,
            )

    def build_target_image(
        self,
        targets_dir: str,
        target: str,
        runtime: RUNTIME_TYPE,
        version: str,
        os_tag: Optional[str] = None,
    ):
        tmp_dir = os.path.join("/tmp", uuid4().hex)
        logger.info(
            f"Building docker image for target {target} ({os.path.join(targets_dir, target)}) in temp dir ({tmp_dir})."
        )
        os.makedirs(tmp_dir, exist_ok=True)
        if not os.path.exists(os.path.join(targets_dir, target)):
            raise TargetBuildException(f"target {target} does not exist.")
        shutil.copytree(os.path.join(targets_dir, target), tmp_dir, dirs_exist_ok=True)
        with open(os.path.join(tmp_dir, "Dockerfile"), "w") as dockerfile:
            dockerfile.write(
                self._get_dockerfile(runtime=runtime, version=version, os_tag=os_tag)
            )
        try:
            _, logs_iter = self.docker_client.images.build(
                path=tmp_dir,
                tag=make_target_image_name(target),
            )
            logs = [l["stream"] for l in logs_iter if "stream" in l]
            self.set_target_build_logs(target, "".join(logs), True)
        except docker.errors.BuildError as e:
            logs = [l["stream"] for l in e.build_log if "stream" in l] + [str(e)]
            self.set_target_build_logs(target, "".join(logs), False)
            raise TargetBuildException("".join(logs))
        finally:
            shutil.rmtree(tmp_dir)

        logger.info(f"Completed building image for target {target}.")

    def get_target_build_logs(self, target: str) -> Tuple[bool, str]:
        results = db_select(
            self.db,
            "SELECT success, build_logs from target_build_logs where target = ?;",
            (target,),
        )
        if len(results) != 1:
            raise BuildLogNotFoundException()
        success, build_logs = results[0]
        return success, build_logs

    def set_target_build_logs(self, target: str, logs: str, success: bool):
        db_execute(
            self.db,
            "INSERT OR REPLACE INTO target_build_logs (target, build_logs, success) VALUES (?, ?, ?)",
            (target, logs, success),
        )

    def _get_dockerfile(
        self, runtime: RUNTIME_TYPE, version: str, os_tag: Optional[str] = None
    ):
        image = runtime if runtime not in ("node-npm", "node-yarn") else "node"
        tag = f"{version}{'' if os_tag is None else f'-{os_tag}'}"
        install = _RUNTIME_SPECIFIC_INSTALL[runtime]
        entrypoint = _RUNTIME_SPECIFIC_ENTRYPOINTS[runtime]
        return f"""
FROM {image}:{tag}

RUN groupadd -r sprinkler-user && useradd -m -r -g sprinkler-user sprinkler-user


WORKDIR /home/sprinkler-user/target
{install}

COPY . /home/sprinkler-user/target

USER sprinkler-user
ENTRYPOINT {entrypoint}
        """


_RUNTIME_SPECIFIC_ENTRYPOINTS = {
    "python": '[ "python3", "-u" ]',
    "node-npm": '[ "node" ]',
    "node-yarn": '[ "node" ]',
}

_RUNTIME_SPECIFIC_INSTALL = {
    "python": """
RUN pip install --user --no-cache-dir sprinkler_util==0.1.3
COPY requirements.txt /home/sprinkler-user/target/requirements.txt
RUN pip install --user --no-cache-dir -r /home/sprinkler-user/target/requirements.txt
    """.strip(),
    "node-yarn": """
COPY yarn.lock /home/sprinkler-user/target/yarn.lock
COPY package.json /home/sprinkler-user/target/package.json
RUN yarn install sprinkler_util@0.1.3
RUN yarn install
    """.strip(),
    "node-npm": """
COPY package-lock.json /home/sprinkler-user/target/package-lock.json
COPY package.json /home/sprinkler-user/target/package.json
RUN npm install sprinkler_util@0.1.3
RUN npm install
    """.strip(),
}



import asyncio
import json
import logging
import sqlite3
from multiprocessing import Process
from typing import Dict, List, Optional, Tuple

import docker
import redis.asyncio as redis
from croniter import croniter

from .constants import (
    _EXECUTE_TASK_INSTANCE_CHANNEL,
    _SCHEDULER_LOGGER,
    _TRIGGER_TYPE_SCHEDULED,
)
from .db import get_db
from .db_interfaces import _CONFIG_TARGETS_DIRECTORY, Config, Images, TaskInstances
from .exceptions import GitException, TargetsDirectoryUnsetException
from .git import Git
from .targets import SprinklerConfig, load_targets
from .utils import now

logger = logging.getLogger(_SCHEDULER_LOGGER)

_SCHEDULER_EXECUTE_LOOP_DELAY = 1
_SCHEDULER_ERROR_LOOP_DELAY = 5

_GIT_PULL_INTERVAL = 30 * 60  # 30 minutes in seconds


class Scheduler:
    redis_client: redis.Redis
    db: sqlite3.Connection
    git: Git
    config: Config
    task_instances: TaskInstances
    images: Images
    targets: List[Tuple[str, SprinklerConfig]]
    build_queue: List[Process]

    last_git_pull: float = 0
    current_git_hash: Optional[str] = None
    recently_started_tasks: Dict[str, float]  # target_task: expiration

    def __init__(self):
        self.redis_client = redis.Redis()
        self.db = get_db()
        self.config = Config(self.db)
        self.git = Git(self.db)
        self.task_instances = TaskInstances(self.db)
        self.images = Images(self.db, docker.from_env())
        self.targets = []
        self.build_queue = []
        self.recently_started_tasks = {}

    async def execute(self):
        self._startup()
        await asyncio.gather(self._queue_loop(), self._execution_loop())

    def _startup(self):
        targets_dir = self.config.get_value(_CONFIG_TARGETS_DIRECTORY)
        if targets_dir is None:
            return
        self.images.build_target_images(targets_dir=targets_dir)
        self.targets = load_targets(targets_dir, include_errors=False)
        self.git_hash = self.git.get_git_hash()
        self.last_git_pull = now()

    async def _queue_loop(self):
        while True:
            await asyncio.sleep(_SCHEDULER_EXECUTE_LOOP_DELAY)
            if len(self.build_queue) == 0:
                continue
            current_process = self.build_queue[0]
            if current_process.pid is None:
                current_process.start()
                continue
            if current_process.is_alive():
                continue
            self.build_queue.pop(0)

    async def _execution_loop(self):
        while True:
            targets_dir = self.config.get_value(_CONFIG_TARGETS_DIRECTORY)
            if targets_dir is None:
                logger.error("TARGETS_DIRECTORY is unset.")
                await asyncio.sleep(_SCHEDULER_ERROR_LOOP_DELAY)
                continue
            if now() - self.last_git_pull > _GIT_PULL_INTERVAL:
                logger.warning("Pulling from remote.")
                try:
                    self.git.pull()
                    self.last_git_pull = now()
                except GitException as e:
                    logger.error(e)
            try:
                git_hash = self.git.get_git_hash()
            except GitException as e:
                logger.error(e)
                await asyncio.sleep(_SCHEDULER_ERROR_LOOP_DELAY)
                continue
            except TargetsDirectoryUnsetException:
                logger.error("TARGETS_DIRECTORY is unset.")
                await asyncio.sleep(_SCHEDULER_ERROR_LOOP_DELAY)
                continue

            if git_hash != self.current_git_hash:
                self.current_git_hash = git_hash
                self.build_queue.append(
                    Process(
                        target=self.images.build_target_images,
                        args=(targets_dir,),
                        daemon=True,
                    )
                )
                self.targets = load_targets(targets_dir, include_errors=False)

            current_time = now()

            cleanup = set()
            for target_task, expiration in self.recently_started_tasks.items():
                if current_time >= expiration:
                    cleanup.add(target_task)
            for target_task in cleanup:
                del self.recently_started_tasks[target_task]

            minute_start_boundary = int(current_time // 60 * 60)
            minute_end_boundary = minute_start_boundary + 60
            for target, config in self.targets:
                for task_config in config.tasks:
                    if task_config.schedule is None:
                        continue
                    if not croniter.match(task_config.schedule, current_time):
                        continue

                    if f"{target}__{task_config.task}" in self.recently_started_tasks:
                        continue

                    existing_task_instances = self.task_instances.get_task_instances(
                        target=[target],
                        task=[task_config.task],
                        trigger_type=[_TRIGGER_TYPE_SCHEDULED],
                        created_at_time_start=minute_start_boundary,
                        created_at_time_end=minute_end_boundary,
                    )
                    if len(existing_task_instances) > 0:
                        continue
                    logger.info(f"Triggering {target} {task_config.task}")
                    self.recently_started_tasks[
                        f"{target}__{task_config.task}"
                    ] = minute_end_boundary

                    asyncio.create_task(
                        self.redis_client.publish(
                            _EXECUTE_TASK_INSTANCE_CHANNEL,
                            json.dumps(
                                {
                                    "target": target,
                                    "task": task_config.task,
                                    "trigger_type": _TRIGGER_TYPE_SCHEDULED,
                                    "response_channel": "null",
                                }
                            ),
                        )
                    )
            await asyncio.sleep(_SCHEDULER_EXECUTE_LOOP_DELAY)



import os
from datetime import datetime


def now():
    return datetime.utcnow().timestamp()


def get_sprinkler_directory():
    return os.path.join(os.path.expanduser("~"), ".sprinkler")


def make_target_image_name(target: str):
    return f"sprinkler-target-image:{target}"



import os
import sqlite3
from typing import List, Literal, Optional, Tuple, Union, overload

from ..db import db_execute, db_select, get_sprinkler_directory
from ..exceptions import InvalidConfigKey


class Config:
    db: sqlite3.Connection

    def __init__(self, db: sqlite3.Connection):
        self.db = db

    def set_value(self, key: str, value: str):
        if key not in _CONFIG_KEYS:
            raise InvalidConfigKey(key)
        db_execute(
            self.db,
            "INSERT OR REPLACE INTO config (key, value) VALUES (?, ?)",
            (key, value),
        )

    @overload
    def get_value(self, key: Literal["GIT_BRANCH", "LOG_LOCATION"]) -> str:
        ...

    @overload
    def get_value(self, key: str) -> Union[str, None]:
        ...

    def get_value(self, key: str):
        key = key.upper()
        if key not in _CONFIG_KEYS:
            raise InvalidConfigKey(key)
        results = db_select(
            self.db, "select value from config where key = ? limit 1;", (key,)
        )
        if len(results) == 0:
            return _CONFIG_DEFAULT_VALUES.get(key)
        return str(results[0][0])

    def ls(self) -> List[Tuple[str, Optional[str]]]:
        return [(key, self.get_value(key)) for key in _CONFIG_KEYS]

    def delete_value(self, key: str):
        key = key.upper()
        if key not in _CONFIG_KEYS:
            raise InvalidConfigKey(key)
        db_execute(self.db, "delete from config where key = ?;", (key,))


_CONFIG_LOG_LOCATION = "LOG_LOCATION"
_CONFIG_TARGETS_DIRECTORY = "TARGETS_DIRECTORY"
_CONFIG_GIT_BRANCH = "GIT_BRANCH"
_CONFIG_FAILURE_TASK_TARGET = "FAILURE_TASK_TARGET"
_CONFIG_FAILURE_TASK = "FAILURE_TASK"
_CONFIG_SSH_KEY = "SSH_KEY"

_CONFIG_KEYS = {
    _CONFIG_LOG_LOCATION,
    _CONFIG_TARGETS_DIRECTORY,
    _CONFIG_GIT_BRANCH,
    _CONFIG_FAILURE_TASK_TARGET,
    _CONFIG_FAILURE_TASK,
    _CONFIG_SSH_KEY,
}

_CONFIG_DEFAULT_VALUES = {
    _CONFIG_LOG_LOCATION: os.path.join(get_sprinkler_directory(), "logs"),
    _CONFIG_GIT_BRANCH: "main",
}



import os
from typing import Optional

from ..constants import _TASK_TYPE_FIRE_AND_FORGET, RUNTIME_TYPE, TASK_TYPE
from ..exceptions import TaskCreationException
from ..utils import make_target_image_name
from .sprinkler_config import (
    InvalidSprinklerConfigException,
    SprinklerFailureTask,
    load_target,
)


class Task:
    target: str
    task: str
    timeout: Optional[int]
    entrypoint_file: str
    image: str
    task_type: TASK_TYPE
    runtime: RUNTIME_TYPE
    version: str
    failure_task: Optional[SprinklerFailureTask] = None

    def __init__(
        self,
        target: str,
        task: str,
        timeout: Optional[int],
        entrypoint_file: str,
        image: str,
        runtime: RUNTIME_TYPE,
        version: str,
        task_type: TASK_TYPE = _TASK_TYPE_FIRE_AND_FORGET,
        failure_task: Optional[SprinklerFailureTask] = None,
    ):
        self.target = target
        self.task = task
        self.timeout = timeout
        self.entrypoint_file = entrypoint_file
        self.image = image
        self.task_type = task_type
        self.runtime = runtime
        self.version = version
        self.failure_task = failure_task


def get_task(targets_dir: Optional[str], target: str, task: str) -> Task:
    if targets_dir is None:
        raise TaskCreationException("TARGETS DIRECTORY is not set")

    try:
        config = load_target(targets_dir, target)

    except InvalidSprinklerConfigException as e:
        raise TaskCreationException(e)

    task_config_list = [t for t in config.tasks if t.task == task]
    if len(task_config_list) == 0:
        raise TaskCreationException(
            f"Failed to find a task with the name {task} in the target {target}"
        )
    if len(task_config_list) > 1:
        raise InvalidSprinklerConfigException(
            f"Found multiple tasks with the name {task} in the target {target}"
        )
    task_config = task_config_list[0]

    failure_task = None
    if config.failure_task is not None:
        failure_task = config.failure_task
    if task_config.failure_task is not None:
        failure_task = task_config.failure_task

    return Task(
        target=target,
        task=task,
        timeout=task_config.timeout,
        entrypoint_file=task_config.entrypoint,
        task_type=task_config.type,
        runtime=config.runtime,
        version=config.version,
        image=make_target_image_name(target),
        failure_task=failure_task,
    )



import sqlite3
from typing import List, Optional

from pydantic import BaseModel

from ..constants import RUNTIME_TYPE, TASK_INSTANCE_STATUS_TYPE, TASK_TYPE, TRIGGER_TYPE
from ..db import db_execute, db_select
from ..exceptions import (
    InvalidTaskInstanceQueryFilter,
    TaskInstanceDoesNotExistException,
)
from ..targets.task import Task
from ..utils import now


class TaskInstanceModel(BaseModel):
    id: str
    target: str
    task: str
    timeout: Optional[int]
    entrypoint_file: str
    task_type: TASK_TYPE
    status: TASK_INSTANCE_STATUS_TYPE
    trigger_type: TRIGGER_TYPE
    created_at_time: float
    runtime: RUNTIME_TYPE
    version: str
    git_hash: str
    start_time: Optional[float]
    end_time: Optional[float]
    log_location: Optional[str]
    container_id: Optional[str]
    response_channel: str
    is_failure_task: bool


class TaskInstances:
    db: sqlite3.Connection

    def __init__(self, db: sqlite3.Connection):
        self.db = db

    def insert_task_instance(
        self,
        id: str,
        task: Task,
        status: TASK_INSTANCE_STATUS_TYPE,
        trigger_type: TRIGGER_TYPE,
        git_hash: str,
        response_channel: str,
        is_error_task: bool,
    ):
        current_timestamp = now()
        db_execute(
            self.db,
            """
INSERT INTO task_instances (
    id,
    target,
    task,
    timeout,
    entrypoint_file,
    task_type,
    status,
    trigger_type,
    created_at_time,
    runtime,
    version,
    git_hash,
    response_channel,
    is_failure_task
)
VALUES
(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?);
        """.strip(),
            (
                id,
                task.target,
                task.task,
                task.timeout,
                task.entrypoint_file,
                task.task_type,
                status,
                trigger_type,
                current_timestamp,
                task.runtime,
                task.version,
                git_hash,
                response_channel,
                is_error_task,
            ),
        )

    def update_task_instance(
        self,
        id: str,
        start_time: Optional[float] = None,
        end_time: Optional[float] = None,
        log_location: Optional[str] = None,
        container_id: Optional[str] = None,
        status: Optional[TASK_INSTANCE_STATUS_TYPE] = None,
    ):
        set_statements = []
        args = []
        if start_time is not None:
            set_statements.append("start_time = ?")
            args.append(start_time)

        if end_time is not None:
            set_statements.append("end_time = ?")
            args.append(end_time)

        if log_location is not None:
            set_statements.append("log_location = ?")
            args.append(log_location)

        if container_id is not None:
            set_statements.append("container_id = ?")
            args.append(container_id)

        if status is not None:
            set_statements.append("status = ?")
            args.append(status)

        if len(set_statements) == 0:
            return
        args.append(id)
        db_execute(
            self.db,
            f"update task_instances SET {' , '.join(set_statements)} where id = ?;",
            tuple(args),
        )

    def get_task_instance(self, id: str):
        results = self.get_task_instances(id=id)
        if len(results) != 1:
            raise TaskInstanceDoesNotExistException(id)
        return results[0]

    def get_task_instances(
        self,
        id: Optional[str] = None,
        limit: int = 100,
        offset: int = 0,
        sort_by: str = "created_at_time",
        sort_ascending: bool = False,
        target: List[str] = [],
        task: List[str] = [],
        task_type: List[str] = [],
        status: List[TASK_INSTANCE_STATUS_TYPE] = [],
        trigger_type: List[TRIGGER_TYPE] = [],
        runtime: List[RUNTIME_TYPE] = [],
        version: List[str] = [],
        git_hash: List[str] = [],
        created_at_time_start: Optional[float] = None,
        created_at_time_end: Optional[float] = None,
        start_time_start: Optional[float] = None,
        start_time_end: Optional[float] = None,
        end_time_start: Optional[float] = None,
        end_time_end: Optional[float] = None,
    ):
        filter_statements = []
        args = []

        if sort_by not in {"created_at_time", "start_time", "end_time"}:
            raise InvalidTaskInstanceQueryFilter(f"sort_by: {sort_by}")

        sort_direction = "asc" if sort_ascending else "desc"

        if len(target) > 0:
            filter_statements.append(_make_in_filter_statement("target", target))
            args += target

        if len(task) > 0:
            filter_statements.append(_make_in_filter_statement("task", task))
            args += task

        if len(task_type) > 0:
            filter_statements.append(_make_in_filter_statement("task_type", task_type))
            args += task_type

        if len(status) > 0:
            filter_statements.append(_make_in_filter_statement("status", status))
            args += status

        if len(trigger_type) > 0:
            filter_statements.append(
                _make_in_filter_statement("trigger_type", trigger_type)
            )
            args += trigger_type

        if len(runtime) > 0:
            filter_statements.append(_make_in_filter_statement("runtime", runtime))
            args += runtime

        if len(version) > 0:
            filter_statements.append(_make_in_filter_statement("version", version))
            args += version

        if len(git_hash) > 0:
            filter_statements.append(_make_in_filter_statement("git_hash", git_hash))
            args += git_hash

        if created_at_time_start is not None:
            filter_statements.append(f"created_at_time >= ?")
            args.append(created_at_time_start)

        if created_at_time_end is not None:
            filter_statements.append(f"created_at_time < ?")
            args.append(created_at_time_end)

        if start_time_start is not None:
            filter_statements.append(f"start_time >= ?")
            args.append(start_time_start)

        if start_time_end is not None:
            filter_statements.append(f"start_time < ?")
            args.append(start_time_end)

        if end_time_start is not None:
            filter_statements.append(f"end_time >= ?")
            args.append(end_time_start)

        if end_time_end is not None:
            filter_statements.append(f"end_time < ?")
            args.append(end_time_end)

        if id is not None:
            filter_statements.append(f"id = ?")
            args.append(id)

        args.append(limit)
        args.append(offset)

        where_clause = (
            ""
            if len(filter_statements) == 0
            else "WHERE " + " AND ".join(filter_statements)
        )
        query = f"""
select id,
    target,
    task,
    timeout,
    entrypoint_file,
    task_type,
    status,
    trigger_type,
    created_at_time,
    runtime,
    version,
    git_hash,
    start_time,
    end_time,
    log_location,
    container_id,
    response_channel,
    is_failure_task
from task_instances
{where_clause}
order by {sort_by} {sort_direction}
limit ? offset ?;
        """.strip()
        results = db_select(
            self.db,
            query,
            tuple(args),
        )

        return [
            TaskInstanceModel(
                id=row[0],
                target=row[1],
                task=row[2],
                timeout=row[3],
                entrypoint_file=row[4],
                task_type=row[5],
                status=row[6],
                trigger_type=row[7],
                created_at_time=row[8],
                runtime=row[9],
                version=row[10],
                git_hash=row[11],
                start_time=row[12],
                end_time=row[13],
                log_location=row[14],
                container_id=row[15],
                response_channel=row[16],
                is_failure_task=row[17],
            )
            for row in results
        ]


def _make_in_filter_statement(column: str, filter_value: list):
    return f"""{column} in ({",".join(["?" for _ in filter_value])})"""



import logging
import os
import sqlite3
import subprocess

from .constants import _GIT_LOGGER
from .db_interfaces import (
    _CONFIG_GIT_BRANCH,
    _CONFIG_SSH_KEY,
    _CONFIG_TARGETS_DIRECTORY,
    Config,
)
from .exceptions import GitException, TargetsDirectoryUnsetException
from .utils import get_sprinkler_directory

logger = logging.getLogger(_GIT_LOGGER)


class Git:
    db: sqlite3.Connection
    config: Config

    def __init__(self, db: sqlite3.Connection):
        self.db = db
        self.config = Config(self.db)

    def _get_targets_dir(self):
        targets_directory = self.config.get_value(_CONFIG_TARGETS_DIRECTORY)
        if targets_directory is None:
            raise TargetsDirectoryUnsetException()
        return targets_directory

    def _ensure_correct_branch(self) -> str:
        self.fetch()
        branch = self.config.get_value(_CONFIG_GIT_BRANCH)
        if self.get_current_branch() != branch:
            logger.info(f"Checking out {branch}.")
            self.checkout(branch)
        return branch

    def _ssh_key_override(self):
        ssh_key = self.config.get_value(_CONFIG_SSH_KEY)
        if ssh_key is None:
            return {}
        return {"GIT_SSH_COMMAND": f"ssh -i {ssh_key} -o IdentitiesOnly=yes"}

    def pull(self):
        branch = self._ensure_correct_branch()
        targets_directory = self._get_targets_dir()
        try:
            return subprocess.check_output(
                ["git", "pull", "origin", branch],
                text=True,
                cwd=targets_directory,
                env=self._ssh_key_override(),
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as e:
            raise GitException(e)

    def get_git_hash(self):
        targets_directory = self._get_targets_dir()
        try:
            git_hash = subprocess.check_output(
                ["git", "rev-parse", "HEAD"], text=True, cwd=targets_directory
            )
        except subprocess.CalledProcessError as e:
            raise GitException(e)
        return git_hash

    def get_current_branch(self):
        targets_directory = self._get_targets_dir()
        try:
            return subprocess.check_output(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"],
                text=True,
                cwd=targets_directory,
            ).strip()
        except subprocess.CalledProcessError as e:
            raise GitException(e)

    def checkout(self, branch: str):
        targets_directory = self._get_targets_dir()
        try:
            return subprocess.check_output(
                ["git", "checkout", branch],
                text=True,
                cwd=targets_directory,
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as e:
            raise GitException(e)

    def fetch(self):
        targets_directory = self._get_targets_dir()
        try:
            return subprocess.check_output(
                ["git", "fetch"],
                text=True,
                cwd=targets_directory,
                env=self._ssh_key_override(),
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as e:
            raise GitException(e)

    def clone(self, dir: str, repo: str):
        try:
            return subprocess.check_output(
                ["git", "clone", repo],
                text=True,
                cwd=dir,
                env=self._ssh_key_override(),
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as e:
            raise GitException(e.output + "\n" + str(e))

    def generate_key(self, force: bool = False):
        sprinkler_dir = get_sprinkler_directory()
        private_key_path = os.path.join(sprinkler_dir, "sprinkler_key")
        public_key_path = os.path.join(sprinkler_dir, "sprinkler_key.pub")
        if (
            os.path.exists(private_key_path) or os.path.exists(public_key_path)
        ) and not force:
            raise GitException("sprinkler_key already exists.")
        if os.path.exists(private_key_path):
            os.remove(private_key_path)
        if os.path.exists(public_key_path):
            os.remove(public_key_path)
        try:
            subprocess.check_output(
                [
                    "ssh-keygen",
                    "-q",
                    "-f",
                    "sprinkler_key",
                    "-t",
                    "ed25519",
                    "-N",
                    "",
                ],
                text=True,
                cwd=sprinkler_dir,
            )
            self.config.set_value(_CONFIG_SSH_KEY, private_key_path)
            return self.get_key()
        except subprocess.CalledProcessError as e:
            raise GitException(e)

    def get_key(self):
        sprinkler_dir = get_sprinkler_directory()
        public_key_path = os.path.join(sprinkler_dir, "sprinkler_key.pub")
        if not os.path.exists(public_key_path):
            raise GitException("Key does not exist.")
        with open(public_key_path, "r") as f:
            return f.read().strip()



import asyncio
import json
import logging
import os
import re
import sqlite3
from typing import Optional
from uuid import uuid4

import docker.errors
import redis.asyncio as redis
import smart_open
from docker import DockerClient
from docker.models.containers import Container

from .constants import (
    _EXECUTE_TASK_INSTANCE_CHANNEL,
    _SPRINKLER_BODY_ENV_VAR,
    _TASK_EXECUTE_LOOP_DELAY_FAST,
    _TASK_EXECUTE_LOOP_DELAY_SLOW,
    _TASK_INSTANCE_LOGGER,
    _TASK_INSTANCE_STATUS_COMPLETE,
    _TASK_INSTANCE_STATUS_ERROR,
    _TASK_INSTANCE_STATUS_RUNNING,
    _TASK_INSTANCE_STATUS_TIMEOUT,
    _TASK_INSTANCE_STATUS_WAITING,
    _TASK_TYPE_CALL_AND_RESPONSE,
    _TRIGGER_TYPE_WEBHOOK,
    TASK_INSTANCE_STATUS_TYPE,
    TRIGGER_TYPE,
)
from .db_interfaces import (
    _CONFIG_FAILURE_TASK,
    _CONFIG_FAILURE_TASK_TARGET,
    _CONFIG_LOG_LOCATION,
    Config,
    Secrets,
    TaskInstanceModel,
    TaskInstances,
)
from .targets import SprinklerFailureTask, Task
from .utils import now

logger = logging.getLogger(_TASK_INSTANCE_LOGGER)


RESPONSE_REGEXP = re.compile(
    r"SPRINKLER\*_\*RESPONSE\*_\*START\*([\s\S]*)\*SPRINKLER\*_\*RESPONSE\*_\*END"
)


class TaskInstance:
    docker_client: DockerClient
    redis_client: redis.Redis
    db: sqlite3.Connection
    config: Config
    secrets: Secrets
    task_instances: TaskInstances
    id: str
    task: Task
    trigger_type: TRIGGER_TYPE
    git_hash: str
    status: TASK_INSTANCE_STATUS_TYPE
    is_failure_task: bool
    response_channel: str

    start_time: Optional[float]
    end_time: Optional[float]
    container: Optional[Container]

    def __init__(
        self,
        task: Task,
        trigger_type: TRIGGER_TYPE,
        git_hash: str,
        docker_client: DockerClient,
        redis_client: redis.Redis,
        db: sqlite3.Connection,
        response_channel: str,
        is_failure_task: bool = False,
        id: Optional[str] = None,
    ):
        self.docker_client = docker_client
        self.redis_client = redis_client
        self.task = task
        self.trigger_type = trigger_type
        self.git_hash = git_hash
        if id is None:
            self.id = TaskInstance._make_task_task_instance_id(
                job=task, trigger_type=trigger_type, git_hash=git_hash
            )
        else:
            self.id = id
        self.response_channel = response_channel
        self.status = _TASK_INSTANCE_STATUS_WAITING
        self.container = None
        self.start_time = None
        self.end_time = None
        self.db = db
        self.is_failure_task = is_failure_task
        self.config = Config(self.db)
        self.secrets = Secrets(self.db)
        self.task_instances = TaskInstances(self.db)

    def store_logs(self, logs: str):
        log_location = self.config.get_value(_CONFIG_LOG_LOCATION)
        log_path = os.path.join(log_location, f"{self.id}.log")
        with smart_open.open(log_path, "w") as log_file:
            log_file.write(logs)
        self.task_instances.update_task_instance(self.id, log_location=log_path)

    def _start_execution(
        self,
        body: str = "",
    ):
        self.task_instances.insert_task_instance(
            self.id,
            self.task,
            self.status,
            self.trigger_type,
            self.git_hash,
            self.response_channel,
            self.is_failure_task,
        )
        logger.info(f"{self.id} | STARTED")
        env = self.secrets.get_secrets_for_execution()
        env[_SPRINKLER_BODY_ENV_VAR] = body
        container = self.docker_client.containers.run(
            self.task.image,
            self.task.entrypoint_file,
            detach=True,
            environment=env,
        )
        if not isinstance(container, Container):
            raise Exception(
                "Starting the docker container did not return a Container object."
            )
        self.container = container
        self.status = _TASK_INSTANCE_STATUS_RUNNING
        self.start_time = now()
        self.task_instances.update_task_instance(
            self.id,
            start_time=self.start_time,
            container_id=self.container.id,
            status=self.status,
        )

    async def _revive(self, task_instance: TaskInstanceModel):
        self.start_time = task_instance.start_time
        self.status = task_instance.status
        self.id = task_instance.id
        if task_instance.container_id is not None:
            try:
                self.container = self.docker_client.containers.get(
                    task_instance.container_id
                )
            except docker.errors.NotFound:
                logger.info(f"{self.id} | REVIVED-DEAD")
                self.end_time = now()
                self.status = _TASK_INSTANCE_STATUS_ERROR
                self.task_instances.update_task_instance(
                    self.id, end_time=self.end_time, status=self.status
                )
                return
        logger.info(f"{self.id} | REVIVED")
        await self._execution_loop()
        await self._after_execution()

    async def _execution_loop(self):
        if self.container is None:
            logger.info(f"{self.id} | CONTAINER ERROR")
            self.status = _TASK_INSTANCE_STATUS_ERROR
            raise Exception("Container object doesn't exist.")
        delay = (
            _TASK_EXECUTE_LOOP_DELAY_FAST
            if self.task.task_type == _TASK_TYPE_CALL_AND_RESPONSE
            else _TASK_EXECUTE_LOOP_DELAY_SLOW
        )

        while True:
            self.container.reload()
            if self.container.status == "exited":
                if self.container.attrs["State"]["ExitCode"] == 0:
                    logger.info(f"{self.id} | COMPLETE")
                    self.status = _TASK_INSTANCE_STATUS_COMPLETE
                    break
                else:
                    logger.info(f"{self.id} | ERROR")
                    self.status = _TASK_INSTANCE_STATUS_ERROR
                    break
            if (
                self.task.timeout is not None
                and self.start_time is not None
                and now() - self.start_time > self.task.timeout
            ):
                logger.info(f"{self.id} | TIMEOUT")
                self.container.kill()
                self.status = _TASK_INSTANCE_STATUS_TIMEOUT
                break
            await asyncio.sleep(delay)

    async def _after_execution(self):
        if self.container is None:
            raise Exception("Container object doesn't exist.")
        self.end_time = now()
        self.task_instances.update_task_instance(
            self.id, end_time=self.end_time, status=self.status
        )
        task_logs = self.container.logs().decode()
        self.container.remove()
        if self.task.task_type == _TASK_TYPE_CALL_AND_RESPONSE:
            response = ""
            response_match = RESPONSE_REGEXP.search(task_logs)
            if response_match is not None:
                response = response_match.group(1)
            await self.redis_client.publish(
                self.response_channel,
                json.dumps({"type": "response", "message": response}),
            )
        if (
            self.status in {_TASK_INSTANCE_STATUS_ERROR, _TASK_INSTANCE_STATUS_TIMEOUT}
            and not self.is_failure_task
        ):
            if self.task.failure_task is not None:
                await self.trigger_failure_task(self.task.failure_task, task_logs)
            else:
                failure_target = self.config.get_value(_CONFIG_FAILURE_TASK_TARGET)
                failure_task = self.config.get_value(_CONFIG_FAILURE_TASK)
                if failure_target is not None and failure_task is not None:
                    await self.trigger_failure_task(
                        SprinklerFailureTask(task=failure_task, target=failure_target),
                        task_logs,
                    )

        self.store_logs(task_logs)

    async def trigger_failure_task(self, failure_task: SprinklerFailureTask, logs: str):
        failure_body = {
            "target": self.task.target,
            "task": self.task.task,
            "id": self.id,
            "runtime": self.task.runtime,
            "version": self.task.version,
            "entrypoint": self.task.entrypoint_file,
            "status": self.status,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "git_hash": self.git_hash,
            "logs": logs,
        }
        await self.redis_client.publish(
            _EXECUTE_TASK_INSTANCE_CHANNEL,
            json.dumps(
                {
                    "target": failure_task.target,
                    "task": failure_task.task,
                    "is_failure_task": True,
                    "trigger_type": _TRIGGER_TYPE_WEBHOOK,
                    "response_channel": "none",
                    "body": json.dumps(failure_body),
                }
            ),
        )

    async def execute(
        self,
        body: str = "",
    ):
        self._start_execution(body=body)
        await self._execution_loop()
        await self._after_execution()

    @staticmethod
    def _make_task_task_instance_id(
        job: Task, trigger_type: TRIGGER_TYPE, git_hash: str
    ):
        return f"{job.target}__{job.task}__{trigger_type}__{round(now())}__{git_hash[:12]}__{uuid4().hex[:12]}"



import sqlite3
from hashlib import sha256
from typing import Optional

from pydantic import BaseModel

from ..db import db_execute, db_select
from ..exceptions import UserAlreadyExistsException, UserDoesNotExistException


class UserModel(BaseModel):
    user_name: str
    admin: bool
    password_hash: Optional[str] = None


class Users:
    db: sqlite3.Connection

    def __init__(self, db: sqlite3.Connection):
        self.db = db

    def create_user(self, username: str, password: str, admin: bool):
        existing_user = db_select(
            self.db, "select 1 from users where user_name = ?;", (username,)
        )
        if len(existing_user) != 0:
            raise UserAlreadyExistsException(username)

        password_hash = sha256(password.encode()).hexdigest()
        db_execute(
            self.db,
            "insert into users (user_name, password_hash, admin) VALUES (?, ?, ?)",
            (username, password_hash, admin),
        )

    def update_user(
        self,
        username: str,
        password: Optional[str] = None,
        admin: Optional[bool] = None,
    ):
        existing_user = db_select(
            self.db, "select 1 from users where user_name = ?;", (username,)
        )
        if len(existing_user) != 1:
            raise UserDoesNotExistException(username)

        if password is not None:
            hashed_password = sha256(password.encode()).hexdigest()
            db_execute(
                self.db,
                "update users set password_hash = ? where user_name = ?;",
                (hashed_password, username),
            )
        if admin is not None:
            db_execute(
                self.db,
                "update users set admin = ? where user_name = ?;",
                (admin, username),
            )

    def delete_user(self, username: str):
        existing_user = db_select(
            self.db, "select 1 from users where user_name = ?;", (username,)
        )
        if len(existing_user) != 1:
            raise UserDoesNotExistException(username)
        db_execute(self.db, "delete from users where user_name = ?;", (username,))

    def get_users(self):
        results = db_select(self.db, "select user_name, admin from users;", None)
        return [UserModel(user_name=result[0], admin=result[1]) for result in results]

    def verify_user_password(self, username: str, password: str):
        password_hash = sha256(password.encode()).hexdigest()
        result = db_select(
            self.db,
            "select 1 from users where user_name = ? and password_hash = ?;",
            (username, password_hash),
        )

        return len(result) == 1



import sqlite3
from hashlib import sha256
from uuid import uuid4

from ..db import db_execute, db_select
from ..exceptions import APIKeyAlreadyExistsException, APIKeyDoesNotExistException


class APIKeys:
    db: sqlite3.Connection

    def __init__(self, db: sqlite3.Connection):
        self.db = db

    def create_api_key(self, api_key_id: str) -> str:
        api_key = uuid4().hex
        hashed_api_key = sha256(api_key.encode()).hexdigest()
        results = db_select(
            self.db, "select 1 from api_keys where api_key_id = ?;", (api_key_id,)
        )
        if len(results) != 0:
            raise APIKeyAlreadyExistsException(api_key_id)

        db_execute(
            self.db,
            "insert into api_keys (api_key_id, api_key_hash) values (?, ?);",
            (api_key_id, hashed_api_key),
        )
        return api_key

    def delete_api_key(self, api_key_id: str):
        results = db_select(
            self.db, "select 1 from api_keys where api_key_id = ?;", (api_key_id,)
        )
        if len(results) == 0:
            raise APIKeyDoesNotExistException(api_key_id)
        db_execute(self.db, "delete from api_keys where api_key_id = ?;", (api_key_id,))

    def get_api_keys(self):
        results = db_select(self.db, "select api_key_id from api_keys;", None)
        return [result[0] for result in results]

    def verify_api_key(self, api_key: str):
        hashed_api_key = sha256(api_key.encode()).hexdigest()
        results = db_select(
            self.db, "select 1 from api_keys where api_key_hash = ?;", (hashed_api_key,)
        )
        return len(results) == 1



from .api_keys import APIKeys
from .config import (
    _CONFIG_FAILURE_TASK,
    _CONFIG_FAILURE_TASK_TARGET,
    _CONFIG_GIT_BRANCH,
    _CONFIG_KEYS,
    _CONFIG_LOG_LOCATION,
    _CONFIG_SSH_KEY,
    _CONFIG_TARGETS_DIRECTORY,
    Config,
)
from .images import Images
from .secrets import Secrets
from .task_instances import TaskInstanceModel, TaskInstances
from .users import UserModel, Users



import asyncio
import json
import logging
import os
import shutil
from typing import Optional
from uuid import uuid4

import click
import docker
import redis
import smart_open

from sprinkler import __version__

from .constants import (
    _EXECUTE_TASK_INSTANCE_CHANNEL,
    _TASK_INSTANCE_STATUS_COMPLETE,
    _TASK_INSTANCE_STATUS_ERROR,
    _TASK_INSTANCE_STATUS_RUNNING,
    _TASK_INSTANCE_STATUS_TIMEOUT,
    _TASK_INSTANCE_STATUS_WAITING,
    _TASK_TYPE_CALL_AND_RESPONSE,
    _TRIGGER_TYPE_WEBHOOK,
)
from .db import db_select, get_db, get_db_file, init_db
from .db_interfaces import (
    _CONFIG_GIT_BRANCH,
    _CONFIG_TARGETS_DIRECTORY,
    APIKeys,
    Config,
    Images,
    Secrets,
    Users,
)
from .exceptions import (
    APIKeyAlreadyExistsException,
    APIKeyDoesNotExistException,
    BuildLogNotFoundException,
    InvalidConfigKey,
    InvalidSprinklerConfigException,
    TargetBuildException,
    UserAlreadyExistsException,
    UserDoesNotExistException,
)
from .git import Git, GitException
from .scheduler import Scheduler
from .targets import load_target, load_targets
from .task_manager import TaskManager
from .utils import get_sprinkler_directory


def init_fn():
    os.makedirs(os.path.join(get_sprinkler_directory(), "logs"), exist_ok=True)
    init_db()
    click.echo(f"Initialized SprinklerDB at {get_db_file()}")


@click.group()
def cli():
    """
    Run sprinkler processes, trigger tasks, get/set config and secrets, and more.

    Built by Henry Jones
    """
    logging.basicConfig(
        level=logging.INFO, format="%(levelname)s %(asctime)s %(name)s | %(message)s"
    )
    if not os.path.exists(get_sprinkler_directory()):
        click.secho(
            "Sprinkler has not yet been initialized. Initializing now.", fg="red"
        )
        init_fn()


@cli.command(name="version")
def version():
    """
    Return the version.
    """
    click.echo(f"Sprinkler {__version__}")


@cli.command(name="webserver")
@click.option("--host", default="0.0.0.0")
@click.option("--port", default=8476)
@click.option("--workers", default=4)
@click.option("--dev", default=False, is_flag=True)
def webserver(host: str, port: int, workers: int, dev: bool):
    """
    Run the Sprinkler webserver.
    """
    if dev:
        from .webserver import app

        app.config["development"] = True
        app.run(debug=True, host=host, port=port)
        return
    from .webserver import StandaloneApplication

    kwargs = {"bind": f"{host}:{port}", "workers": workers}
    StandaloneApplication(kwargs).run()


@cli.command(name="init")
def init_fn_command():
    """Initialize the Sprinkler environment."""
    init_fn()


@cli.command()
def task_manager():
    """Starts the Sprinkler Task Manager"""
    click.echo("Starting Sprinkler Task Manager")
    tm = TaskManager()
    asyncio.run(tm.execute())


@cli.command()
def scheduler():
    """Starts the Sprinkler Scheduler"""
    click.echo("Starting Sprinkler Scheduler")
    sc = Scheduler()
    asyncio.run(sc.execute())


@cli.group()
def git():
    """
    Targets git related actions.
    """


@git.command("pull")
def git_pull():
    """
    Pulls from remote.
    """
    git = Git(get_db())
    try:
        click.echo(git.pull())
    except GitException as e:
        raise click.ClickException(str(e))


@git.command("checkout")
@click.argument("branch")
def git_checkout(branch: str):
    """
    Checks out the given branch and updates the config.
    """
    db = get_db()
    git = Git(db)
    config = Config(db)
    try:
        config.set_value(_CONFIG_GIT_BRANCH, branch)
        click.echo(git.pull())
    except GitException as e:
        raise click.ClickException(str(e))


@git.command("init")
@click.argument("directory")
@click.argument("git-remote")
@click.option("--branch", default=None)
@click.option("--force", default=False, is_flag=True)
def git_init(directory: str, git_remote: str, branch: Optional[str], force: bool):
    """
    Initializes the target directory with a remote git repo.
    """
    if not os.path.exists(directory):
        raise click.ClickException(f"{directory} does not exist.")
    if not os.path.isdir(directory):
        raise click.ClickException(f"{directory} is not a directory.")
    d = git_remote.split("/")[-1].replace(".git", "")
    targets_dir = os.path.join(directory, d)

    db = get_db()
    git = Git(db)
    config = Config(db)
    try:
        if os.path.exists(targets_dir):
            if force:
                click.secho(f"Deleting {targets_dir}", fg="red")
                shutil.rmtree(targets_dir)
            else:
                raise click.ClickException(
                    click.style(
                        f"{targets_dir} already exists. Use --force to remove it.",
                        fg="yellow",
                    )
                )
        click.echo(f"Setting TARGETS_DIRECTORY to {targets_dir}.")
        config.set_value(_CONFIG_TARGETS_DIRECTORY, targets_dir)
        if branch is not None:
            click.echo(f"Setting GIT_BRANCH to {branch}.")
            config.set_value(_CONFIG_GIT_BRANCH, branch)
        click.echo(git.clone(directory, git_remote))
        click.echo(git.pull())
    except GitException as e:
        raise click.ClickException(str(e))


@git.command("generate-key")
@click.option("-f", "--force", default=False, is_flag=True)
def git_generate_key(force: bool):
    """
    Generate a ssh-key to be used by sprinkler.
    """
    db = get_db()
    git = Git(db)
    try:
        click.echo(git.generate_key(force=force))
        db.close()
    except GitException as e:
        raise click.ClickException(str(e))


@git.command("get-key")
def git_get_key():
    """
    Get the sprinkler public key.
    """
    db = get_db()
    git = Git(db)

    try:
        click.echo(git.get_key())
        db.close()
    except GitException as e:
        raise click.ClickException(str(e))


@cli.group()
def users():
    """
    Create, delete, or update users.
    """


@users.command(name="create")
@click.argument("username")
@click.argument("password")
@click.option("-a", "--admin", default=False, is_flag=True)
def create_user(username: str, password: str, admin: bool):
    """
    Create a new user.
    """
    users = Users(get_db())
    try:
        users.create_user(username, password, admin)
    except UserAlreadyExistsException:
        raise click.ClickException(
            f"A user with the username {username} already exists."
        )
    click.echo(f"Created user {username}.")


@users.command(name="delete")
@click.argument("username")
def delete_user(username: str):
    """
    Deletes a user.
    """
    users = Users(get_db())
    try:
        users.delete_user(username)
    except UserDoesNotExistException:
        raise click.ClickException(
            f"A user with the username {username} does not exist."
        )
    click.echo(f"Deleted user {username}.")


@users.command("update-password")
@click.argument("username")
@click.argument("password")
def update_user_password(username: str, password: str):
    """
    Updates a user's password.
    """
    users = Users(get_db())
    try:
        users.update_user(username, password=password)
    except UserDoesNotExistException:
        raise click.ClickException(
            f"A user with the username {username} does not exist."
        )
    click.echo(f"Updated user {username} password.")


@users.command("promote")
@click.argument("username")
def promote_user_password(username: str):
    """
    Makes a user an admin.
    """
    users = Users(get_db())
    try:
        users.update_user(username, admin=True)
    except UserDoesNotExistException:
        raise click.ClickException(
            f"A user with the username {username} does not exist."
        )
    click.echo(f"Promoted user {username}.")


@users.command("demote")
@click.argument("username")
def demote_user_password(username: str):
    """
    Revoke a user's admin status.
    """
    users = Users(get_db())
    try:
        users.update_user(username, admin=False)
    except UserDoesNotExistException:
        raise click.ClickException(
            f"A user with the username {username} does not exist."
        )
    click.echo(f"Demoted user {username}.")


@users.command("ls")
def ls_users():
    """
    Lists users and their admin status.
    """
    users = Users(get_db())
    results = users.get_users()
    for user in results:
        click.echo(
            f"{user.user_name} {click.style('admin', fg='green') if user.admin else click.style('not-admin', fg='red')}"
        )


@cli.group("api-keys")
def api_keys():
    """
    Create, list, or delete API Keys.
    """
    pass


@api_keys.command("create")
@click.argument("api-key-id")
def create_api_key(api_key_id: str):
    """
    Create an API Key. Returns the API key.

    Save it in a safe place as it cannot be retrieved again.
    """
    api_keys = APIKeys(get_db())
    try:
        api_key = api_keys.create_api_key(api_key_id)
    except APIKeyAlreadyExistsException:
        raise click.ClickException(
            f"An API Key already exists with the id {api_key_id}"
        )
    click.echo(api_key)


@api_keys.command("delete")
@click.argument("api-key-id")
def delete_api_key(api_key_id: str):
    """
    Deletes an API Key.
    """
    api_keys = APIKeys(get_db())
    try:
        api_keys.delete_api_key(api_key_id)
    except APIKeyDoesNotExistException:
        raise click.ClickException(
            f"An API Key does not exists with the id {api_key_id}"
        )
    click.echo(f"Deleted API Key {api_key_id}.")


@api_keys.command("ls")
def ls_api_keys():
    """
    Lists API Keys.
    """
    api_keys = APIKeys(get_db())
    results = api_keys.get_api_keys()
    for api_key_id in results:
        click.echo(api_key_id)


@cli.group()
def config():
    """
    Get, set, or list config values.
    """
    pass


@config.command(name="ls")
def config_ls():
    """
    List config options.
    """
    c = Config(get_db())
    for key, value in c.ls():
        click.echo(f"{key}  {value}")


@config.command(name="set")
@click.argument("key")
@click.argument("value")
def config_set(key: str, value: str):
    """
    Set a config option.
    """
    c = Config(get_db())
    try:
        c.set_value(key, value)
        click.echo(f"SET {key} {value}")
    except InvalidConfigKey as e:
        raise click.ClickException(f"Invalid config key `{e}`")


@config.command(name="get")
@click.argument("key")
def config_get(key: str):
    """
    Get a config value.
    """
    c = Config(get_db())
    try:
        v = c.get_value(key)
        click.echo(v)
    except InvalidConfigKey as e:
        raise click.ClickException(f"Invalid config key `{e}`")


@config.command(name="delete")
@click.argument("key")
def config_delete(key: str):
    c = Config(get_db())
    try:
        c.delete_value(key)
        click.echo(f"Deleted {key}")
    except InvalidConfigKey as e:
        raise click.ClickException(f"Invalid config key `{e}`")


@cli.group()
def secrets():
    """
    Get, set, delete, and list secrets.
    """
    pass


@secrets.command(name="set")
@click.argument("key")
@click.argument("value")
def secrets_set(key: str, value: str):
    """
    Set a secret value.
    """
    s = Secrets(get_db())
    s.set_secret(key, value)
    click.echo(f"SET {key}")


@secrets.command(name="get")
@click.argument("key")
def secrets_get(key: str):
    """
    Get a secret value.
    """
    s = Secrets(get_db())
    secret_value = s.get_secret(key)
    if secret_value is None:
        raise click.ClickException(f"No secret found with the key `{key}`.")
    click.echo(secret_value)


@secrets.command("delete")
@click.argument("key")
def secrets_delete(key: str):
    s = Secrets(get_db())
    s.delete_secret(key)
    click.echo(f"Deleted {key}.")


@secrets.command(name="ls")
def secrets_ls():
    """
    List set secrets.
    """
    s = Secrets(get_db())
    keys = s.get_secret_keys()
    for key in keys:
        click.echo(key)


@cli.group()
def targets():
    """
    Build, list, and describe targets.
    """
    pass


@targets.command(name="build-logs")
@click.argument("target")
def target_build_logs(target: str):
    """
    Retrieve the latest build logs for a target.
    """
    images = Images(get_db(), None)
    try:
        success, build_logs = images.get_target_build_logs(target)
    except BuildLogNotFoundException:
        raise click.ClickException(f"No build logs found for the target {target}.")
    click.echo(
        f"Build status for {click.style(target, bold=True)}: {click.style('SUCCESS', fg='green') if success else click.style('FAILED', fg='red')}"
    )
    click.echo("=========Build Logs==========")
    click.echo(build_logs)


@targets.command(name="build")
@click.argument("target")
@click.option("-l", "--logs", default=False, is_flag=True)
def target_build(target: str, logs: bool):
    """
    Build the image for a target.
    """
    db = get_db()
    targets_dir = Config(db).get_value(_CONFIG_TARGETS_DIRECTORY)
    if targets_dir is None:
        raise click.ClickException(f"{_CONFIG_TARGETS_DIRECTORY} is unset.")
    try:
        config = load_target(targets_dir, target)
    except InvalidSprinklerConfigException as e:
        raise click.ClickException(str(e))
    images = Images(db, docker.from_env())
    try:
        images.build_target_image(
            targets_dir,
            target,
            config.runtime,
            config.version,
            config.os,
        )
    except TargetBuildException as e:
        raise click.ClickException(f"Build Failed:\n{e}")
    if not logs:
        return
    try:
        _, build_logs = images.get_target_build_logs(target)
    except BuildLogNotFoundException:
        raise click.ClickException(f"No build logs found for the target {target}.")
    click.echo("=========Build Logs==========")
    click.echo(build_logs)


@targets.command(name="describe")
@click.argument("target")
def target_describe(target: str):
    """
    Describe a target.
    """
    targets_dir = Config(get_db()).get_value(_CONFIG_TARGETS_DIRECTORY)
    if targets_dir is None:
        raise click.ClickException(f"{_CONFIG_TARGETS_DIRECTORY} is unset.")

    try:
        config = load_target(targets_dir, target)
    except InvalidSprinklerConfigException as e:
        raise click.ClickException(str(e))

    click.echo(
        f"{click.style(target, bold=True)} {click.style(config.runtime, fg='yellow')} {click.style(config.version, fg='green')} {'' if config.os is None else click.style(config.os, fg='cyan')}"
    )

    for task in config.tasks:
        click.echo("=" * 40)
        click.secho(task.task, bold=True)
        click.echo(f"{task.entrypoint} {task.type}")
        if task.schedule:
            click.echo(f"Schedule: {task.schedule}")
        if task.timeout:
            click.echo(f"Timeout: {task.timeout} seconds.")


@targets.command(name="ls")
def targets_ls():
    """
    List targets and their validity.
    """
    targets_dir = Config(get_db()).get_value(_CONFIG_TARGETS_DIRECTORY)
    if targets_dir is None:
        raise click.ClickException(f"{_CONFIG_TARGETS_DIRECTORY} is unset.")
    for target, config in load_targets(targets_dir=targets_dir, include_errors=True):
        if isinstance(config, str):
            click.echo(f'{target} {click.style("invalid", fg="red")} {config}')
            continue
        click.echo(f'{target} {click.style("valid", fg="green")}')


@cli.group()
def tasks():
    """
    Trigger tasks.
    """
    pass


@tasks.command(name="trigger")
@click.argument("target")
@click.argument("task")
@click.option("--body")
def tasks_trigger(target: str, task: str, body: str = ""):
    """
    Trigger a task.
    """
    redis_client = redis.Redis()
    response_channel = f"sprinkler-response-channel:{uuid4().hex}"
    with redis_client.pubsub(ignore_subscribe_messages=True) as pub_sub:
        pub_sub.subscribe(response_channel)
        num_subscribers = redis_client.publish(
            _EXECUTE_TASK_INSTANCE_CHANNEL,
            json.dumps(
                {
                    "target": target,
                    "task": task,
                    "trigger_type": _TRIGGER_TYPE_WEBHOOK,
                    "response_channel": response_channel,
                    "body": json.dumps(body),
                }
            ),
        )
        if num_subscribers == 0:
            raise click.ClickException(
                f"Nobody was listening. Are you sure that the Sprinkler Task Manager is running?"
            )
        for message in pub_sub.listen():
            message_data = json.loads(message["data"].decode())
            if message_data.get("type") == "receipt":
                if message_data.get("success") is not True:
                    raise click.ClickException(message_data.get("message"))
                click.echo(message_data.get("message"))
                if message_data.get("task_type") == _TASK_TYPE_CALL_AND_RESPONSE:
                    continue
                return

            if message_data.get("type") == "response":
                click.echo(message_data.get("message"))
                return
            raise click.ClickException(
                "Did not get a valid response from the task manager."
            )
    raise click.ClickException("Internal Service Error")


@cli.command(name="logs")
@click.argument("task-instance-id")
@click.option("-f", "--follow", default=False, is_flag=True)
def task_instance_logs(task_instance_id: str, follow: bool = False):
    """
    Get logs for a task instance.
    """
    db = get_db()
    results = db_select(
        db,
        "select status, log_location, container_id from task_instances where id = ?;",
        (task_instance_id,),
    )
    if len(results) != 1:
        raise click.ClickException(f"No task found with id {task_instance_id}")
    status, log_location, container_id = results[0]
    if status == _TASK_INSTANCE_STATUS_WAITING:
        click.echo(f"The task {task_instance_id} is waiting to start.")
        return
    if status == _TASK_INSTANCE_STATUS_WAITING:
        status_txt = click.style(status, fg="grey")
    elif status == _TASK_INSTANCE_STATUS_RUNNING:
        status_txt = click.style(status, fg="yellow")
    elif status == _TASK_INSTANCE_STATUS_COMPLETE:
        status_txt = click.style(status, fg="green")
    elif status in (_TASK_INSTANCE_STATUS_ERROR, _TASK_INSTANCE_STATUS_TIMEOUT):
        status_txt = click.style(status, fg="red")
    else:
        status_txt = status
    click.echo(f"{task_instance_id} | {status_txt}")
    click.echo("=" * 20)
    if status == _TASK_INSTANCE_STATUS_RUNNING and container_id is not None:
        docker_client = docker.from_env()
        try:
            container = docker_client.containers.get(container_id)
            if follow:
                for line in container.logs(stream=True):
                    click.echo(line.strip())
            else:
                click.echo(container.logs())
        except docker.errors.NotFound:
            raise click.ClickException(f"Container {container_id} not found.")
    elif log_location is not None:
        with smart_open.open(log_location, "r") as f:
            click.echo(f.read())


if __name__ == "__main__":
    cli()



import asyncio
import json
import logging
import sqlite3
from typing import Dict, Optional, Union

import docker
import redis.asyncio as redis

from .constants import (
    _EXECUTE_TASK_INSTANCE_CHANNEL,
    _TASK_INSTANCE_STATUS_RUNNING,
    _TASK_MANAGER_LOGGER,
    TASK_TYPE,
)
from .db import get_db
from .db_interfaces import _CONFIG_TARGETS_DIRECTORY, Config, TaskInstances
from .git import Git
from .targets import Task, get_task
from .task_instance import TRIGGER_TYPE, TaskInstance

logger = logging.getLogger(_TASK_MANAGER_LOGGER)


class TaskManager:
    docker_client: docker.DockerClient
    redis_client: redis.Redis
    db: sqlite3.Connection
    config: Config
    git: Git
    tasks: Dict[str, TaskInstance]

    def __init__(self):
        self.docker_client = docker.from_env()
        self.redis_client = redis.Redis()
        self.db = get_db()
        self.config = Config(self.db)
        self.git = Git(self.db)

    async def _revive(self):
        task_instances = TaskInstances(self.db)
        orphan_task_instances = task_instances.get_task_instances(
            limit=10_000_000, status=[_TASK_INSTANCE_STATUS_RUNNING]
        )
        for orphan in orphan_task_instances:
            task_instance = self._create_task_instance(
                orphan.target,
                orphan.task,
                orphan.trigger_type,
                orphan.response_channel,
                is_failure_task=orphan.is_failure_task,
            )

            asyncio.create_task(task_instance._revive(orphan))

    async def execute(self):
        await self._revive()
        async with self.redis_client.pubsub() as pub_sub:
            await pub_sub.subscribe(_EXECUTE_TASK_INSTANCE_CHANNEL)
            logger.info(f"Subscribed to {_EXECUTE_TASK_INSTANCE_CHANNEL}")
            while True:
                message = await pub_sub.get_message(ignore_subscribe_messages=True)
                if message is None:
                    await asyncio.sleep(0.01)
                    continue
                decoded_message = message["data"].decode()
                try:
                    message_data = json.loads(decoded_message)
                except json.JSONDecodeError:
                    logger.error(
                        f"Failed to parse subscription message: '{decoded_message}'"
                    )
                    continue
                if "response_channel" not in message_data:
                    logger.error("No response channel given in the message.")
                try:
                    task_instance = self._create_task_instance(
                        message_data["target"],
                        message_data["task"],
                        message_data["trigger_type"],
                        message_data["response_channel"],
                        message_data.get("is_failure_task", False),
                    )
                except Exception as e:
                    logger.error(f"Failed to create task instance: {e}")
                    await self.send_trigger_response(
                        message_data["response_channel"],
                        False,
                        message=f"Failed to create task instance: {e}",
                    )
                    continue
                asyncio.create_task(
                    task_instance.execute(
                        body=message_data.get("body", ""),
                    )
                )
                await self.send_trigger_response(
                    message_data["response_channel"],
                    True,
                    job_type=task_instance.task.task_type,
                    message=task_instance.id,
                )

    async def send_trigger_response(
        self,
        response_channel: str,
        success: bool,
        job_type: Optional[TASK_TYPE] = None,
        message: Optional[str] = None,
    ):
        response_message: Dict[str, Union[str, bool]] = {
            "success": success,
            "type": "receipt",
        }
        if message is not None:
            response_message["message"] = message
        if job_type is not None:
            response_message["job_type"] = job_type

        await self.redis_client.publish(response_channel, json.dumps(response_message))

    def _create_task_instance(
        self,
        target: str,
        task: str,
        trigger_type: TRIGGER_TYPE,
        response_channel: str,
        is_failure_task: bool,
    ):
        _task = get_task(self.config.get_value(_CONFIG_TARGETS_DIRECTORY), target, task)
        git_hash = self.git.get_git_hash()
        return TaskInstance(
            task=_task,
            trigger_type=trigger_type,
            docker_client=self.docker_client,
            db=self.db,
            redis_client=self.redis_client,
            git_hash=git_hash,
            is_failure_task=is_failure_task,
            response_channel=response_channel,
        )



import json
import logging
import os
import shutil
from uuid import uuid4

import docker
import redis
import smart_open
from flask import Flask, jsonify, request, send_from_directory, session
from flask_cors import CORS

from sprinkler import __version__

from .constants import (
    _EXECUTE_TASK_INSTANCE_CHANNEL,
    _SPRINKLER_SECRET_KEY_ENVIRONMENT_VARIABLE,
    _TASK_INSTANCE_STATUS_RUNNING,
    _TASK_INSTANCE_STATUS_WAITING,
    _TASK_TYPE_CALL_AND_RESPONSE,
    _TRIGGER_TYPE_WEBHOOK,
    _WEBSERVER_LOGGER,
)
from .db import get_db
from .db_interfaces import (
    _CONFIG_GIT_BRANCH,
    _CONFIG_TARGETS_DIRECTORY,
    APIKeys,
    Config,
    Images,
    Secrets,
    TaskInstances,
    Users,
)
from .exceptions import (
    BuildLogNotFoundException,
    GitException,
    InvalidConfigKey,
    InvalidSecretName,
    InvalidSprinklerConfigException,
    TargetBuildException,
    TargetsDirectoryUnsetException,
    TaskInstanceDoesNotExistException,
)
from .git import Git
from .targets import load_target, load_targets
from .utils import get_sprinkler_directory

logger = logging.getLogger(_WEBSERVER_LOGGER)
app = Flask(__name__)
app.config["development"] = False
CORS(app)
app.secret_key = os.environ.get(_SPRINKLER_SECRET_KEY_ENVIRONMENT_VARIABLE, "sprinkler")


_WHITELISTED_PATHS = {"/login", "/access_check"}


@app.before_request
def before_requests():
    if app.config["development"]:
        return
    if request.path[-1:] == "/":
        return
    if request.path in _WHITELISTED_PATHS:
        return
    if request.path[:8] == "/assets/":
        return

    user = session.get("user")
    if user is not None:
        return

    api_key = request.headers.get("x-api-key")
    if api_key is None:
        return "Access Denied", 403

    db = get_db()
    api_keys = APIKeys(db)
    is_valid = api_keys.verify_api_key(api_key)
    db.close()
    if is_valid:
        return
    return "Access Denied", 403


@app.get("/access_check")
def access_check():
    if app.config["development"]:
        return jsonify({"is_authorized": True, "type": "user", "user_name": "JOHN_DOE"})
    user = session.get("user")
    if user is not None:
        return jsonify({"is_authorized": True, "type": "user", "user_name": user})
    api_key = request.headers.get("x-api-key")
    if api_key is not None:
        db = get_db()
        api_keys = APIKeys(db)
        is_valid_api_key = api_keys.verify_api_key(api_key)
        db.close()
        if is_valid_api_key:
            return jsonify({"is_authorized": True, "type": "api_key"})

    return jsonify({"is_authorized": False})


@app.post("/login")
def login():
    body = request.get_json(force=True)
    if "user_name" not in body or "password" not in body:
        return jsonify({"success": False})

    db = get_db()
    users = Users(db)
    valid_user = users.verify_user_password(body["user_name"], body["password"])
    db.close()

    if valid_user:
        session["user"] = body["user_name"]
        return jsonify({"success": True})
    return jsonify({"success": False})


@app.get("/assets/<path:path>")
def assets(path: str):
    build_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), ".build", "assets"
    )
    return send_from_directory(build_path, path)


@app.get("/")
@app.get("/<path:path>/")
def index(path: str = ""):
    build_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        ".build",
    )
    return send_from_directory(build_path, "index.html")


@app.post("/target/<string:target>/task/<string:task>/trigger")
def trigger_target_job(target: str, task: str):
    body = request.get_data(as_text=True)
    redis_client = redis.Redis()
    response_channel = f"sprinkler-response-channel:{uuid4().hex}"
    with redis_client.pubsub(ignore_subscribe_messages=True) as pub_sub:
        pub_sub.subscribe(response_channel)
        subscribers = redis_client.publish(
            _EXECUTE_TASK_INSTANCE_CHANNEL,
            json.dumps(
                {
                    "target": target,
                    "task": task,
                    "trigger_type": _TRIGGER_TYPE_WEBHOOK,
                    "response_channel": response_channel,
                    "body": json.dumps(body),
                }
            ),
        )
        if subscribers == 0:
            return (
                jsonify(
                    "Nothing was listening. Are you sure sprinkler task-manager was listening?"
                ),
                500,
            )
        for message in pub_sub.listen():
            message_data = json.loads(message["data"].decode())
            if message_data.get("type") == "receipt":
                if message_data.get("success") is not True:
                    return jsonify(message_data.get("message")), 404
                if message_data.get("job_type") == _TASK_TYPE_CALL_AND_RESPONSE:
                    continue
                return jsonify(f"Started {message_data.get('message')} successfully")

            if message_data.get("type") == "response":
                return message_data.get("message"), 200
            return jsonify("Did not get a valid response from the task manager."), 500
    return jsonify("Internal Service Error"), 500


@app.get("/config")
def get_config():
    db = get_db()
    c = Config(db)
    configuration = {key: value if value else "" for key, value in c.ls()}
    db.close()
    return jsonify(configuration)


@app.post("/config/<string:key>")
def set_config(key: str):
    value = request.get_data(as_text=True)
    db = get_db()
    c = Config(db)
    try:
        c.set_value(key, value)
    except InvalidConfigKey:
        db.close()
        return f"Invalid config key {key}", 428
    db.close()
    return get_config()


@app.delete("/config/<string:key>")
def delete_config(key: str):
    db = get_db()
    c = Config(db)
    try:
        c.delete_value(key)
    except InvalidConfigKey:
        db.close()
        return f"Invalid config key {key}", 428
    db.close()
    return get_config()


@app.get("/secrets")
def get_secrets():
    db = get_db()
    s = Secrets(db)
    secrets = s.get_secret_keys()
    db.close()
    return jsonify(secrets)


@app.post("/secrets/<string:key>")
def set_secret(key):
    value = request.get_data(as_text=True)
    db = get_db()
    s = Secrets(db)
    try:
        s.set_secret(key, value)
    except InvalidSecretName as e:
        return jsonify(f"The key was invalid: {str(key)}"), 428
    db.close()
    return get_secrets()


@app.delete("/secrets/<string:key>")
def delete_secret(key):
    db = get_db()
    s = Secrets(db)
    s.delete_secret(key)
    db.close()
    return get_secrets()


@app.get("/targets")
def get_targets():
    targets_dir = Config(get_db()).get_value(_CONFIG_TARGETS_DIRECTORY)
    if targets_dir is None:
        return jsonify([])
    targets = []
    for target, config in load_targets(targets_dir=targets_dir, include_errors=True):
        if isinstance(config, str):
            targets.append({"target": target, "error": config})
            continue
        targets.append({"target": target, "error": None})
    return jsonify(targets)


@app.get("/targets/<string:target>")
def get_target_detail(target: str):
    targets_dir = Config(get_db()).get_value(_CONFIG_TARGETS_DIRECTORY)
    if targets_dir is None:
        return jsonify(f"{_CONFIG_TARGETS_DIRECTORY} is unset."), 500
    try:
        config = load_target(targets_dir, target)
    except InvalidSprinklerConfigException as e:
        return jsonify({"error": str(e)})

    return jsonify(config.dict())


@app.post("/targets/<string:target>/build")
def build_target(target: str):
    db = get_db()
    targets_dir = Config(db).get_value(_CONFIG_TARGETS_DIRECTORY)
    if targets_dir is None:
        return jsonify(f"{_CONFIG_TARGETS_DIRECTORY} is unset."), 500

    try:
        config = load_target(targets_dir, target)
    except InvalidSprinklerConfigException as e:
        return jsonify(f"{target} is invalid: {e}"), 500
    images = Images(db, docker.from_env())
    try:
        images.build_target_image(
            targets_dir=targets_dir,
            target=target,
            runtime=config.runtime,
            version=config.version,
            os_tag=config.os,
        )
    except TargetBuildException as e:
        return jsonify({"success": False, "logs": str(e)})
    finally:
        db.close()
    return jsonify({"success": True})


@app.get("/targets/<string:target>/build-logs")
def build_logs_target(target: str):
    db = get_db()
    images = Images(db, None)
    try:
        success, build_logs = images.get_target_build_logs(target)
    except BuildLogNotFoundException:
        return jsonify({"logs": None})

    return jsonify({"build_success": success, "logs": build_logs})


@app.get("/task_instance")
def get_task_instances():
    db = get_db()
    task_instances = TaskInstances(db)
    results = task_instances.get_task_instances(
        limit=request.args.get("limit", 100, type=int),
        offset=request.args.get("offset", 0, type=int),
        sort_by=request.args.get("sort_by", "created_at_time"),
        sort_ascending="sort_ascending" in request.args,
        target=request.args.getlist("target"),
        task=request.args.getlist("task"),
        status=request.args.getlist("status"),  # type: ignore
        trigger_type=request.args.getlist("trigger_type"),  # type: ignore
        runtime=request.args.getlist("runtime"),  # type: ignore
        version=request.args.getlist("version"),
        git_hash=request.args.getlist("git_hash"),
        created_at_time_start=request.args.get("created_at_time_start", type=float),
        created_at_time_end=request.args.get("created_at_time_end", type=float),
        start_time_start=request.args.get("start_time_start", type=float),
        start_time_end=request.args.get("start_time_end", type=float),
        end_time_start=request.args.get("end_time_start", type=float),
        end_time_end=request.args.get("end_time_end", type=float),
    )
    db.close()
    return jsonify([result.dict() for result in results])


@app.get("/task_instance/<string:task_instance_id>")
def get_task_instance(task_instance_id: str):
    db = get_db()
    task_instances = TaskInstances(db)
    try:
        task_instance = task_instances.get_task_instance(task_instance_id)
    except TaskInstanceDoesNotExistException:
        return jsonify(f"task instance {task_instance_id} not found"), 404
    return jsonify(task_instance.dict())


@app.get("/task_instance/<string:task_instance_id>/logs")
def get_task_instance_logs(task_instance_id: str):
    db = get_db()
    task_instances = TaskInstances(db)
    try:
        task_instance = task_instances.get_task_instance(task_instance_id)
    except TaskInstanceDoesNotExistException:
        return jsonify(f"task instance {task_instance_id} not found"), 404

    db.close()
    if task_instance.status == _TASK_INSTANCE_STATUS_WAITING:
        return jsonify("")
    if (
        task_instance.status == _TASK_INSTANCE_STATUS_RUNNING
        and task_instance.container_id is not None
    ):
        docker_client = docker.from_env()
        try:
            container = docker_client.containers.get(task_instance.container_id)
            return jsonify(container.logs().decode())
        except docker.errors.NotFound:
            return jsonify(f"Container {task_instance.container_id} not found."), 404
    elif task_instance.log_location is not None:
        with smart_open.open(task_instance.log_location, "r") as f:
            return jsonify(f.read())
    return jsonify("No Logs Found."), 500


@app.get("/version")
def version():
    return jsonify(__version__)


@app.get("/git/hash")
def get_git_hash():
    db = get_db()
    git = Git(db)
    try:
        git_hash = git.get_git_hash()
    except GitException as e:
        return jsonify(f"GIT exception: {e}"), 500
    except TargetsDirectoryUnsetException:
        return jsonify(f"{_CONFIG_TARGETS_DIRECTORY} is unset."), 500
    db.close()
    return jsonify(git_hash)


@app.post("/git/pull")
def git_pull():
    db = get_db()
    git = Git(db)
    try:
        git.pull()
        git_hash = git.get_current_branch()
    except GitException as e:
        return jsonify(f"GIT exception: {e}"), 500
    except TargetsDirectoryUnsetException:
        return jsonify(f"{_CONFIG_TARGETS_DIRECTORY} is unset."), 500
    db.close()
    return jsonify(git_hash)


@app.post("/git/checkout")
def git_checkout():
    branch = request.get_data(as_text=True)
    db = get_db()
    git = Git(db)
    config = Config(db)
    config.set_value(_CONFIG_GIT_BRANCH, branch)
    try:
        git.pull()
        git_hash = git.get_current_branch()
    except GitException as e:
        return jsonify(f"GIT exception: {e}"), 500
    except TargetsDirectoryUnsetException:
        return jsonify(f"{_CONFIG_TARGETS_DIRECTORY} is unset."), 500
    db.close()
    return jsonify(git_hash)


@app.post("/git/generate_key")
def git_generate_key():
    force = "force" in request.args
    db = get_db()
    git = Git(db)
    try:
        public_key = git.generate_key(force=force)
    except GitException as e:
        return jsonify(f"GIT exception: {e}"), 500
    return jsonify(public_key)


@app.get("/git/get_key")
def git_get_key():
    db = get_db()
    git = Git(db)
    try:
        public_key = git.get_key()
    except GitException as e:
        return jsonify(f"GIT exception: {e}"), 500
    return jsonify(public_key)


@app.post("/git/init")
def git_init():
    data = request.get_json(force=True)
    directory = data.get("directory", get_sprinkler_directory())
    git_remote = data.get("git_remote")
    if git_remote is None:
        return jsonify("Must specify git_remote."), 500
    branch = data.get("branch")
    force = data.get("force")

    if not os.path.exists(directory):
        return jsonify(f"{directory} does not exist."), 404
    if not os.path.isdir(directory):
        return jsonify(f"{directory} is not a directory."), 404
    d = git_remote.split("/")[-1].replace(".git", "")
    targets_dir = os.path.join(directory, d)

    db = get_db()
    git = Git(db)
    config = Config(db)
    logs = ""
    try:
        if os.path.exists(targets_dir):
            if force:
                logger.warning(f"Deleting {targets_dir}")
                shutil.rmtree(targets_dir)
            else:
                return jsonify(
                    f"{targets_dir} already exists. Specify force=True to remove it.",
                )

        if branch is not None:
            config.set_value(_CONFIG_GIT_BRANCH, branch)
        logs += git.clone(directory, git_remote) + "\n"
        config.set_value(_CONFIG_TARGETS_DIRECTORY, targets_dir)
        logs += git.pull()
    except GitException as e:
        return jsonify(logs + "\n" + str(e)), 500
    return jsonify(logs)


@app.get("/api_keys")
def get_api_keys():
    db = get_db()
    api_keys = APIKeys(db)
    keys = api_keys.get_api_keys()
    db.close()
    return jsonify(keys)


@app.post("/api_keys/<string:key>")
def create_api_keys(key: str):
    db = get_db()
    api_keys = APIKeys(db)
    api_key = api_keys.create_api_key(key)
    db.close()
    return jsonify(api_key)


@app.delete("/api_keys/<string:key>")
def delete_api_key(key: str):
    db = get_db()
    api_keys = APIKeys(db)
    api_key = api_keys.delete_api_key(key)
    db.close()
    return jsonify(api_key)


### LAUNCHING GUNICORN SERER STUFF ##############
from gunicorn.app.base import BaseApplication


class StandaloneApplication(BaseApplication):
    def __init__(self, options=None):
        self.options = options or {}
        self.application = app
        super().__init__()

    def init(self, parser, opts, args):
        pass

    def load_config(self):
        for key, value in self.options.items():
            if key in self.cfg.settings and value is not None:
                self.cfg.set(key.lower(), value)

    def load(self):
        return self.application



import os

from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="s7r",
    author="Henry Jones",
    author_email="henryivesjones@gmail.com",
    url="https://github.com/henryivesjones/sprinkler",
    description="An easy to use job runner.",
    packages=["sprinkler", "sprinkler.db_interfaces", "sprinkler.targets"],
    package_dir={
        "sprinkler": "sprinkler",
        "sprinkler.db_interfaces": "sprinkler/db_interfaces",
        "sprinkler.targets": "sprinkler/targets",
    },
    package_data={
        "": [".build/**"],
        "sprinkler": [
            "py.typed",
        ],
        "sprinkler.db_interfaces": ["py.typed"],
        "sprinkler.targets": ["py.typed"],
    },
    include_package_data=True,
    long_description=read("README.md"),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
    ],
)



import os
from typing import List, Literal, Optional, Tuple, Union, overload

from pydantic import BaseModel, ValidationError
from yaml import YAMLError, load

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from ..constants import (
    _CONFIG_YAML_FILENAME,
    _TASK_TYPE_FIRE_AND_FORGET,
    RUNTIME_TYPE,
    TASK_TYPE,
)
from ..exceptions import InvalidSprinklerConfigException


class SprinklerFailureTask(BaseModel):
    target: str
    task: str


class SprinklerTaskConfig(BaseModel):
    task: str
    entrypoint: str
    schedule: Optional[str]
    timeout: Optional[int]
    type: TASK_TYPE = _TASK_TYPE_FIRE_AND_FORGET
    failure_task: Optional[SprinklerFailureTask] = None


class SprinklerConfig(BaseModel):
    tasks: List[SprinklerTaskConfig]
    runtime: RUNTIME_TYPE
    version: str
    failure_task: Optional[SprinklerFailureTask] = None
    os: Optional[str] = None


@overload
def load_targets(
    targets_dir: str, include_errors: Literal[False]
) -> List[Tuple[str, SprinklerConfig]]:
    ...


@overload
def load_targets(
    targets_dir: str, include_errors: Literal[True]
) -> List[Tuple[str, Union[SprinklerConfig, str]]]:
    ...


def load_targets(targets_dir: str, include_errors: bool = False):
    targets = []
    for target in os.listdir(targets_dir):
        if target == ".git":
            continue
        target_dir = os.path.join(targets_dir, target)
        if not os.path.isdir(target_dir):
            continue
        try:
            config = load_target(targets_dir, target)
        except InvalidSprinklerConfigException as e:
            if include_errors:
                targets.append((target, str(e)))
            continue
        targets.append((target, config))
    return targets


def load_target(targets_dir: str, target: str) -> SprinklerConfig:
    target_dir = os.path.join(targets_dir, target)
    if not os.path.isdir(target_dir):
        raise InvalidSprinklerConfigException("Target doesn't exist.")

    config_file = os.path.join(target_dir, _CONFIG_YAML_FILENAME)
    if not os.path.isfile(config_file):
        raise InvalidSprinklerConfigException(f"No {_CONFIG_YAML_FILENAME}")

    with open(config_file, "r") as f:
        try:
            config_yaml = load(f, Loader)
        except YAMLError:
            raise InvalidSprinklerConfigException("Invalid YAML")

    try:
        config = SprinklerConfig(**config_yaml)
    except ValidationError as e:
        raise InvalidSprinklerConfigException(f"YAML did not pass validation. {e}")
    return config
